package com.yourkit.probes.builtin;

import com.yourkit.probes.*;
import com.yourkit.runtime.Callback;
import org.jetbrains.annotations.Nullable;

public final class Servlets {
  public static final String TABLE_NAME = "JSP/Servlet";

  private static final class RequestTable extends Table {
    private final StringColumn myURI = new StringColumn("URI");
    private final StringColumn myQuery = new StringColumn("Query");

    public RequestTable() {
      super(Servlets.class, TABLE_NAME, Table.MASK_FOR_LASTING_EVENTS);
    }
  }
  private static final RequestTable T_REQUEST = new RequestTable();

  @MethodPattern(
    {
      "*:service(javax.servlet.http.HttpServletRequest, javax.servlet.http.HttpServletResponse) void",
      "-javax.servlet.http.HttpServlet:*(*)",
      "-org.apache.jasper.*:*(*)"
    }
  )
  @InstanceOf("javax.servlet.http.HttpServlet")
  public static final class Servlet_service_Probe {
    public static int onEnter(
      @This final Object servlet,
      @Param(1) final Object request
    ) {
      // ignore calls to Servlet's service() for jsp pages
      try {
        final Class<?> jspPageClass = Class.forName("javax.servlet.jsp.HttpJspPage", false, request.getClass().getClassLoader());
        if (jspPageClass.isInstance(servlet)) {
          return Table.NO_ROW;
        }
      }
      catch (final Throwable ignored) {
      }

      return createRequestRow(request);
    }

    public static void onExit(
      @OnEnterResult final int rowIndex,
      @ThrownException @Nullable final Throwable e
    ) {
      T_REQUEST.closeRow(rowIndex, e);
    }
  }

  @MethodPattern("*:_jspService(javax.servlet.http.HttpServletRequest, javax.servlet.http.HttpServletResponse) void")
  @InstanceOf("javax.servlet.jsp.HttpJspPage")
  public static final class HttpJspPage_jspService_Probe {
    public static int onEnter(
      @Param(1) final Object request
    ) {
      return createRequestRow(request);
    }

    public static void onExit(
      @OnEnterResult final int rowIndex,
      @ThrownException @Nullable final Throwable e
    ) {
      T_REQUEST.closeRow(rowIndex, e);
    }
  }

  @MethodPattern("*:doFilter(javax.servlet.ServletRequest, javax.servlet.ServletResponse, javax.servlet.FilterChain)")
  @InstanceOf("javax.servlet.Filter")
  public static final class Filter_doFilter_Probe {
    public static int onEnter(
      @Param(1) final Object request
    ) {
      return createRequestRow(request);
    }

    public static void onExit(
      @OnEnterResult final int rowIndex,
      @ThrownException @Nullable final Throwable e
    ) {
      T_REQUEST.closeRow(rowIndex, e);
    }
  }

  private static int createRequestRow(final Object request) {
    final int rowIndex = T_REQUEST.createRow();
    if (Table.shouldIgnoreRow(rowIndex)) {
      // optimization
      return Table.NO_ROW;
    }

    final String includeRequestURI = getRequestAttribute(request, "javax.servlet.include.request_uri");

    final String requestURI;
    final String query;
    if (includeRequestURI == null || includeRequestURI.isEmpty()) {
      requestURI = getRequestURI(request);
      query = getQuery(request);
    }
    else {
      requestURI = includeRequestURI;
      query = getRequestAttribute(request, "javax.servlet.include.query_string");
    }

    T_REQUEST.myURI.setValue(rowIndex, requestURI);
    T_REQUEST.myQuery.setValue(rowIndex, query);

    return rowIndex;
  }

  /**
   * This method uses reflection to avoid class-loading issues
   */
  private static String getQuery(final Object request) {
    if (request == null) {
      return "<unknown>";
    }
    try {
      return (String)Callback.callObjectMethod0(request.getClass(), request, "getQueryString", "()Ljava/lang/String;");
    }
    catch (final Throwable ignored) {
      return getQuery(tryGetWrappedRequest(request));
    }
  }

  /**
   * This method uses reflection to avoid class-loading issues
   */
  static String getRequestURI(final Object request) {
    if (request == null) {
      return "<unknown>";
    }
    try {
      return (String)Callback.callObjectMethod0(request.getClass(), request, "getRequestURI", "()Ljava/lang/String;");
    }
    catch (final Throwable ignored) {
      return getRequestURI(tryGetWrappedRequest(request));
    }
  }

  /**
   * This method uses reflection to avoid class-loading issues
   */
  private static String getRequestAttribute(final Object request, final String attributeName) {
    if (request == null) {
      return "<unknown>";
    }
    try {
      final Class<?> requestClass = request.getClass();
      final String result = (String)Callback.callObjectMethod1(
        requestClass, request, "getAttribute", "(Ljava/lang/String;)Ljava/lang/Object;", attributeName
      );
      if (result != null && result.startsWith("//")) {
        return result.substring(1);
      }
      return result;
    }
    catch (final Throwable ignored) {
      return getRequestAttribute(tryGetWrappedRequest(request), attributeName);
    }
  }

  /**
   * This method uses reflection to avoid class-loading issues
   */
  private static Object tryGetWrappedRequest(final Object requestWrapper) {
    try {
      return Callback.callObjectMethod0(requestWrapper.getClass(), requestWrapper, "getRequest", "()Ljavax/servlet/ServletRequest;");
    }
    catch (final Throwable ignored) {
      return null;
    }
  }
}
